/* misc.c
 * Various functions, mostly for logging.
 */

#include "weedlog.h"

#include <stdarg.h>

static char argstr[MAX_LOG_BUFFER];

#ifdef NEEDSTRERROR
#if !defined(__GLIBC__) || (__GLIBC__ < 2)
char *strerror(int error) { 
  char *undef = "Undefined error";
  if(error > sys_nerr)
    return undef;
  return sys_errlist[error];
}
#endif
#endif

/* Stuff for logging.
 */
void print_log(const char *logmsg, ...) {
  va_list vl;

  va_start(vl, logmsg);
  vsnprintf(argstr, MAX_LOG_BUFFER, logmsg, vl);
  printb_log(FALSE, argstr);
  va_end(vl);
  memset(argstr, 0, sizeof(argstr));
}

void print_log_err(const char *logerr, ...) {
  va_list vl;

  va_start(vl, logerr);
  vsnprintf(argstr, MAX_LOG_BUFFER, logerr, vl);
  printb_log(TRUE, argstr);
  va_end(vl);
  memset(argstr, 0, sizeof(argstr));
}

void printb_log(int doexit, char *logmsg) {
  if(HasLSysL(logmode)) {
    syslog(LOG_NOTICE, "%s", logmsg);
  }
  if(HasLFile(logmode)) {
    time_t t;
    struct tm *lt;
    char *tempstr, *timestring;
#ifdef USETHREADS
    if(filelock) {
      while(filelock) usleep(1);
    }
    filelock = 1;
#endif
    if((tempstr = (char*)malloc(PATH_MAX)) < 0) {
      fprintf(stderr, "%s[%i]: malloc: %s\n", progname, wlpid, strerror(errno));
      return;
    }
    t = time(NULL);
    lt = localtime(&t);
    if((timestring = (char*)malloc(24)) < 0) {
      fprintf(stderr, "%s[%i]: malloc: %s\n", progname, wlpid, strerror(errno));
      return;
    }
    if((strftime(timestring, 24, "%b %d %Y %T", lt)) < 0) {
      fprintf(stderr, "%s[%i]: strftime: %s", progname, wlpid, strerror(errno));
    }
    snprintf(tempstr, PATH_MAX, "%s %s[%i]: %s", timestring, progname, wlpid, logmsg);
    if((write(outfile, tempstr, wl_psize(tempstr))) < 0) {
      fprintf(stderr, "%s[%i]: write(outfile): %s", progname, wlpid, strerror(errno));
    }
    if(tempstr) free(tempstr);
    if(timestring) free(timestring);
#ifdef USETHREADS
    filelock = 0;
#endif
  }
  if(HasLSOut(logmode)) {
    time_t t;
    struct tm *lt;
    char *timestring;
#ifdef USETHREADS
    if(stdlock) {
      while(stdlock) usleep(1);
    }
    stdlock = 1;
#endif
    t = time(NULL);
    lt = localtime(&t);
    if((timestring = (char*)malloc(24)) < 0) {
      fprintf(stderr, "%s[%i]: malloc: %s\n", progname, wlpid, strerror(errno));
      return;
    }
    if((strftime(timestring, 24, "%b %d %Y %T", lt)) < 0) {
      fprintf(stderr, "%s[%i]: strftime: %s", progname, wlpid, strerror(errno));
    }
    fprintf(stdout, "%s %s[%i]: %s", timestring, progname, wlpid, logmsg);
    if(timestring) free(timestring);
#ifdef USETHREADS
    stdlock = 0;
#endif
  }

  if(doexit)
    exit(1);
}

void print_log_add(char *logmsg, ...) {
  va_list vl;

  va_start(vl, logmsg);
  vsnprintf(argstr, MAX_LOG_BUFFER, logmsg, vl);
  printb_log_add(argstr);
  va_end(vl);
  memset(argstr, 0, sizeof(argstr));
}

void printb_log_add(char *logmsg) {
  if(logmode & LOGMODE_SYSL) {
    syslog(LOG_NOTICE, "%s", logmsg);
  }
  if(logmode & LOGMODE_FILE) {
    char *tempstr;
    if((tempstr = (char*)malloc(PATH_MAX)) < 0) {
      fprintf(stderr, "malloc: %s\n", strerror(errno));
    }
    snprintf(tempstr, PATH_MAX, "%s", logmsg);
    if((write(outfile, tempstr, wl_psize(tempstr))) < 0) {
      fprintf(stderr, "write(outfile): %s", strerror(errno));
    }
    if(tempstr) free(tempstr);
  }
  if(logmode & LOGMODE_STDOUT) {
    fprintf(stdout, "%s", logmsg);
  }
}

/* Stuff forlooking up port and host addresses.
 */
char * hostlookup(unsigned long int in) {
  static char host[65];
  struct in_addr i;
  struct hostent *he;

  i.s_addr = in;
  he = gethostbyaddr((char *)&i, sizeof(i),AF_INET);
  if(he == NULL)
    strncpy(host, inet_ntoa(i), sizeof(host));
  else
    strncpy(host, he->h_name, sizeof(host));
  return host;
}

char * servlookup(unsigned short port, int proto) {
  struct servent *se = (struct servent *)NULL;
  static char buff[128];
   
  if(proto == 6)
    se = getservbyport(port, "tcp");
  else if(proto == 17)
    se = getservbyport(port, "udp");
  if(se == NULL)
    snprintf(buff, sizeof(buff), "%d", ntohs(port));
  else
    snprintf(buff, sizeof(buff), "%s", se->s_name);
  return buff;
}

