/* sockets.c
 * Stuff for dealing with sockets such as creating them and reading them.
 */

#include "weedlog.h"

void setup_sockets(int socktypes, wl_verb verb) {
  int types = socktypes;

  if(types > SOCKTYPE_ALL) types = SOCKTYPE_ALL;

whilestart:
  while(types) {
    int s = -1, proto = 0, ptype = 0, pverb = 0, tvl = 0;
    if(TCPIsSet(types)) {
      proto = 6;
      ptype = SOCKTYPE_TCP;
      if(verb.tcp) pverb = verb.tcp;
      ClrTCP(types);
    } else if(UDPIsSet(types)) {
      proto = 17;
      ptype = SOCKTYPE_UDP;
      if(verb.udp) pverb = verb.udp;
      ClrUDP(types);
    } else if(ICMPIsSet(types)) {
      proto = 1;
      ptype = SOCKTYPE_ICMP;
      if(verb.icmp) pverb = verb.icmp;
      ClrICMP(types);
    } else if(IGMPIsSet(types)) {
      proto = 2;
      ptype = SOCKTYPE_IGMP;
      if(verb.igmp) pverb = verb.igmp;
      ClrIGMP(types);
    } else {
      goto whilestart;
    }
    if((s = socket(AF_INET, SOCK_RAW, proto)) < 0) {
      print_log_err("Error creating protocol %i socket. Errno: %s\n", proto, strerror(errno));
    }
    setsock_nonblock(s,TRUE);
    if(pverb) tvl = pverb;
    else if(verb.global) tvl = verb.global;
    else tvl = 0;
    addsocket(s, ptype, tvl);
  }
}

int addsocket(int sock, int type, int verbl) {
  wl_socket *socktoadd, *tmp = socketlist;

  if((socktoadd = (wl_socket*)malloc(sizeof(*socktoadd))) < 0) {
    print_log("malloc: %s", strerror(errno));
    return -1;
  }
  socktoadd->next = (wl_socket*)NULL;
  socktoadd->prev = (wl_socket*)NULL;
  socktoadd->fd = sock;
  socktoadd->type = type;
  socktoadd->verbose = verbl;
  if(socketlist == (wl_socket*)NULL) {
    socketlist = socktoadd;
    return 1;
  } else {
    while(tmp) {
      if(!tmp->next) {
        tmp->next = socktoadd;
	socktoadd->prev = tmp;
	return 1;
      } else tmp = tmp->next;
    }
  }
  return 0;
}

int setsock_nonblock(int fd, int dontblock) {
  int opts;

  if(dontblock) {
    if((opts = fcntl(fd,F_GETFL)) < 0) {
      print_log("fcntl (F_GETFL): %s", strerror(errno));
      return -1;
    }
    opts |= O_NONBLOCK;
    if(fcntl(fd,F_SETFL,opts) < 0) {
      print_log("fcntl (F_SETFL): %s", strerror(errno));
      return -1;
    }
  } else {
    if((opts = fcntl(fd,F_GETFL)) < 0) {
      print_log("fcntl (F_GETFL): %s", strerror(errno));
      return -1;
    }
    opts &= ~O_NONBLOCK;
    if(fcntl(fd,F_SETFL,opts) < 0) {
      print_log("fcntl (F_SETFL): %s", strerror(errno));
      return -1;
    }
  }
  return 0;
}

int check_for_packet() {
  u_char *pbuff;
  wl_socket *checksocks = socketlist;
  struct sockaddr from;
  int amountread = 0, fromlen, fdproto = 0;

  if(!checksocks)
    return 0;

  if((pbuff = (u_char*)malloc(PACKET_BUF_SIZE)) < 0){
    print_log("check_for_packet: malloc: %s", strerror(errno));
    return -1;
  }

  fromlen = sizeof(from);
  while(checksocks) {
    if(checksocks->fd) {
      GetProto(checksocks->type,fdproto);
      if((amountread = recvfrom(checksocks->fd, (u_char*)pbuff,
	  PACKET_BUF_SIZE, 0, &from, &fromlen)) < 0) {
	if(errno) {
	  if(errno != 11) { /* EAGAIN a.k.a. EWOULDBLOCK */
	    print_log("error: Received error %i (%s) while attempting to read"
		" from file descriptor %i (proto: %i).\n", errno,
		strerror(errno), checksocks->fd, fdproto);
	  } else {
	    if(debug >= 5)
	      print_log("debug: Got nothing on file descriptor %i (proto: %i).\n",
		  checksocks->fd, fdproto);
	  }
	}
      } else {
	int protosupported = 0;
	struct parse_args *pargs;
#ifdef USETHREADS
	pthread_t parse_thread;
#endif
	void *(*parse_func)(void*) = NULL;
	if(debug >= 5) {
	  print_log("debug: Got something on file descriptor %i (proto: %i).\n",
	      checksocks->fd, fdproto);
	  print_log("debug: read %i, fromlen == %i, from.data "
	      "== %s, from.sa_family == %i\n", amountread, fromlen,
	      from.sa_data, from.sa_family);
	}
	if((pargs = (struct parse_args *)malloc(sizeof(struct parse_args))) < 0) {
	  print_log("check_for_packet: malloc: %s", strerror(errno));
	  return -1;
	}
	if((pargs->buffer = (u_char*)malloc(PACKET_BUF_SIZE)) < 0){
	  print_log("check_for_packet: malloc: %s", strerror(errno));
	  return -1;
	}
	memcpy(pargs->buffer, pbuff, PACKET_BUF_SIZE);
	pargs->verblev = checksocks->verbose;
	if(ICMPIsSet(checksocks->type)) {
	  protosupported = 1;
	  parse_func = (thread_function_t)parse_icmp;
	} else if(IGMPIsSet(checksocks->type)) {
	  protosupported = 1;
	  parse_func = (thread_function_t)parse_igmp;
	} else if(TCPIsSet(checksocks->type)) {
	  protosupported = 1;
	  parse_func = (thread_function_t)parse_tcp;
	} else if(UDPIsSet(checksocks->type)) {
	  protosupported = 1;
	  parse_func = (thread_function_t)parse_udp;
	} else {
	  protosupported = 0;
	  if(pargs->buffer) free(pargs->buffer);
	  if(pargs) free(pargs);
	}
	if(protosupported && parse_func) {
#ifdef USETHREADS
	  if((pthread_create(&parse_thread, (pthread_attr_t*)NULL,
	      (thread_function_t)parse_func, (void*)pargs)) < 0) {
	    if(debug >= 4)
	      printb_log(FALSE, "debug: check_for_packet: pthread_create failed\n");
	    parse_func((void*)pargs);
	  } else {
	    if(debug >= 4)
	      printb_log(FALSE, "debug: check_for_packet: detaching thread\n");
	    activethreads++;
	    pthread_detach(parse_thread);
	  }
#else /* ! USETHREADS */
	  parse_func((void*)pargs);
#endif
	}
      }
    }
    checksocks = checksocks->next;
  }
  if(pbuff)
    free(pbuff);
  return 0;
}
